@@ -1,21 +1,17 @@
 import os
+import re
 import uuid
 import time
 import socket
 import tempfile
 import logging
-import string
-from random import choice
-from StringIO import StringIO
 
-import traceback
 import paramiko
 from boto.ec2.connection import EC2Connection
 from boto.ec2 import elb
+from boto.ec2 import blockdevicemapping
 from boto.exception import BotoServerError
 from fabric.api import *
-from fabric.contrib import files
-from fabulaws.api import *
 
 logger = logging.getLogger('fabulaws.ec2')
 
@@ -39,7 +35,7 @@ class EC2Service(object):
     def setup(self):
         self.conn = self._connect_ec2()
 
-    def instances(self, filters=None, cls=None, inst_kwargs=None):
+    def instances(self, filters=None, cls=None, inst_kwargs=None, instance_ids=None):
         """
         Return list of all matching reservation instances
         """
@@ -48,7 +44,8 @@ class EC2Service(object):
             filters['instance-state-name'] = 'running'
         cls = cls or EC2Instance
         inst_kwargs = inst_kwargs or {}
-        reservations = self.conn.get_all_instances(filters=filters)
+        reservations = self.conn.get_all_instances(filters=filters,
+                                                   instance_ids=instance_ids)
         results = []
         for reservation in reservations:
             for instance in reservation.instances:
@@ -71,12 +68,24 @@ class EC2Instance(object):
     user = ''
     admin_groups = []
     instance_type = ''
-    security_groups = []
+    security_groups = []  # Names of the instance's security groups
     key_prefix = ''
     ssh_timeout = 5
 
     _saved_contexts = []
 
+    instance_storage = {
+        'm1.small': ['/dev/xvdb'],
+        'm3.medium': ['/dev/xvdb'],
+        'm3.large': ['/dev/xvdb'],
+        'm3.xlarge': ['/dev/xvdb', '/dev/xvdc'],
+        'c3.large': ['/dev/xvdb', '/dev/xvdc'],
+    }
+
+    # instances that have EBS optimization by default
+    # (see http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/EBSOptimized.html)
+    ebs_optimized_regex = re.compile(r'^(c4|m4|d2)')
+
     def __init__(self, access_key_id=None, secret_access_key=None,
                  terminate=False, placement=None, tags=None, instance_id=None,
                  instance=None):
@@ -110,6 +119,17 @@ class EC2Instance(object):
             self.conn = self._connect_ec2()
         self.elb_conn = self._connect_elb()
 
+    def get_security_groups_for_launch_configuration(self):
+        """
+        Return a list of the instance's security group identifiers in
+        the format needed for a launch configuration.  This is names for
+        EC-2 classic, and IDs for modern accounts.
+        """
+        if getattr(self, 'instance', False) and getattr(self.instance, 'vpc_id', False):
+            # It has a VPC so it's modern
+            return [group.id for group in self.instance.groups]
+        return self.security_groups
+
     def _connect_ec2(self):
         logger.debug('Connecting to EC2')
         return EC2Connection(self._key_id, self._secret)
@@ -156,12 +176,31 @@ class EC2Instance(object):
         else:
             logger.info('Creating EC2 instances')
             image = self.conn.get_image(ami)
+            if image is None:
+                raise ValueError('AMI {0} not found'.format(ami))
             key_name = self.key and self.key.name or None
-            res = image.run(key_name=key_name,
-                            security_groups=self.security_groups,
-                            instance_type=self.instance_type,
-                            placement=placement,
-                            min_count=count, max_count=count)
+            bdm = blockdevicemapping.BlockDeviceMapping()
+            for i, dev in enumerate(self.instance_storage.get(self.instance_type, [])):
+                eph_name = 'ephemeral{}'.format(i)
+                bdm[dev] = blockdevicemapping.BlockDeviceType(ephemeral_name=eph_name)
+            ebs_optimized = bool(self.ebs_optimized_regex.match(self.instance_type))
+            if hasattr(env, 'subnets'):
+                # We're in a VPC, so provide the 'subnet_id' and 'security_group_ids'.
+                extra_kwargs = {
+                    'subnet_id': env.subnets[placement],
+                    'security_group_ids': self.security_groups,
+                }
+            else:
+                # We're NOT in a VPC, so provide the 'placement' and 'security_groups'.
+                extra_kwargs = {
+                    'placement': placement,
+                    'security_groups': self.security_groups,
+                }
+            res = self.conn.run_instances(image.id, key_name=key_name,
+                                          instance_type=self.instance_type,
+                                          block_device_map=bdm,
+                                          min_count=count, max_count=count,
+                                          ebs_optimized=ebs_optimized, **extra_kwargs)
             time.sleep(5) # wait for AWS to catch up
             created = True
         for inst in res.instances:
@@ -199,11 +238,11 @@ class EC2Instance(object):
             try:
                 key = self.key_file and self.key_file.name or env.key_filename
                 user = self.user and self.user or env.user
-                ssh.connect(instance.public_dns_name, allow_agent=False,
-                            look_for_keys=False, username=user,
+                ssh.connect(getattr(instance, getattr(env, 'ec2_attr_for_ssh', 'public_dns_name')),
+                            allow_agent=False, look_for_keys=False, username=user,
                             key_filename=key, timeout=self.ssh_timeout)
                 break
-            except (EOFError, socket.error, paramiko.SSHException), e:
+            except (EOFError, socket.error, paramiko.SSHException) as e:
                 logger.debug('Error connecting ({0}); retrying in {1} '
                              'seconds'.format(e, wait))
                 times += 1
@@ -228,9 +267,10 @@ class EC2Instance(object):
         if self.user:
             logger.debug('Setting env.user = "{0}"'.format(self.user))
             env.user = self.user
+        host_string = getattr(self.instance, getattr(env, 'ec2_attr_for_ssh', 'public_dns_name'))
         logger.debug('Setting env.host_string = "{0}"'
-                     ''.format(self.instance.public_dns_name))
-        env.host_string = self.instance.public_dns_name
+                     ''.format(host_string))
+        env.host_string = host_string
         env.current_server = self
 
     def _restore_context(self):
@@ -286,8 +326,11 @@ class EC2Instance(object):
         """
         Adds this instance to the specified load balancer.
         """
-        # ensure that the load balancer accepts traffic to this AV
-        self.elb_conn.enable_availability_zones(elb_name, [self._placement])
+        try:
+            # ensure that the load balancer accepts traffic to this AV
+            self.elb_conn.enable_availability_zones(elb_name, [self._placement])
+        except BotoServerError:
+            logger.info('Failed to call enable_availability_zones(). IGNORE this error if using a VPC!')
         return self.elb_conn.register_instances(elb_name, [self.instance.id])
 
     def remove_from_elb(self, elb_name):
@@ -352,7 +395,11 @@ class EC2Instance(object):
         logger.info('Waiting for image to enter "available" state...')
         while image.update() == 'pending':
             time.sleep(2)
-        assert image.update() == 'available'
+        status = image.update()
+        while status != 'available':
+            logger.info('Unexpected image status after pending: %s...waiting a bit longer...', status)
+            time.sleep(2)
+            status = image.update()
         logger.info('Image creation finished.')
         return image
 
@@ -392,6 +439,13 @@ class EC2Instance(object):
             self.key_file.close()
             self.key_file = None
 
+    def terminate(self):
+        """
+        Terminates this instance in EC2.
+        """
+        self._terminate = True
+        self.cleanup()
+
     def add_tags(self, tags):
         """
         Associate specified tags with instance
@@ -403,7 +457,7 @@ class EC2Instance(object):
     @property
     def hostname(self):
         if self.instance:
-            return self.instance.public_dns_name
+            return getattr(self.instance, getattr(env, 'ec2_attr_for_ssh', 'public_dns_name'))
         else:
             raise ValueError('No instance has been created yet, or the '
                              'instance has already been destroyed.')
